/*
Copyright 2012 Nathanael Hübbe
Contact: nathanael.huebbe@informatik.uni-hamburg.de

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <hdf5.h>
#include <stdbool.h>

const int kDimensionCount = 3;
const int kVariableCount = 40;
const int kRecordCount = 20;
const hsize_t kDimensionSizes[] = {1, 20000, 100};

// Use -r | --reduced to get reduced output (only the naked throughput in MiB/s is printed).
int main(int argc, char** argv) {
	int myProcId, procCount;
	hid_t fileId, spaceId, transferPlist, memspace, datasetCreationPList, setIds[kVariableCount];
	hsize_t myStartCoord, reducedDimensionSizes[kDimensionCount], dimensionMaxSizes[kDimensionCount], curDimensionSizes[kDimensionCount];
	hsize_t startCoords[kDimensionCount];
	long i, floatCount;
	char name[100];	//I know this is bad, but I can't use glibc on the blizzard, so there's no asprintf available.
	struct timeval startTime, endTime;
	bool reducedOutput = false, collectiveCalls = false;
	
	MPI_Init(&argc, &argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &myProcId);
	MPI_Comm_size(MPI_COMM_WORLD, &procCount);
	
	for(i = 1; i < argc; i++) {
		if(!strcmp(argv[i], "-r") || !strcmp(argv[i], "--reduced")) reducedOutput = true;
		else if(!strcmp(argv[i], "-c") || !strcmp(argv[i], "--collective")) collectiveCalls = true;
		else {
			if(!myProcId) {
				printf("Unknown argument: %s\n", argv[i]);
				printf("Supported options:\n");
				printf("\t-c | --collective Use collective calls for output\n");
				printf("\t-r | --reduced Print only the naked throughput in MiB/s\n");
			}
			return -1;
		}
	}

	// Get the area that is to be written by this process.
	for(i = 0; i < kDimensionCount; i++) reducedDimensionSizes[i] = kDimensionSizes[i];
	myStartCoord = (myProcId*(uint64_t)kDimensionSizes[1])/procCount;
	reducedDimensionSizes[1] = ((myProcId + 1)*(uint64_t)kDimensionSizes[1])/procCount - myStartCoord;

	MPI_Barrier(MPI_COMM_WORLD);
	if(gettimeofday(&startTime, NULL)) return -2;
	
	// Create the file.
	{
		hid_t plist_id = H5Pcreate(H5P_FILE_ACCESS);
		if(0 > plist_id) return -3;
		if(0 > H5Pset_fapl_mpio(plist_id, MPI_COMM_WORLD, MPI_INFO_NULL)) return -4;
		if(0 > (fileId = H5Fcreate("/work/ku0598/bench.h5", H5F_ACC_TRUNC, H5P_DEFAULT, plist_id))) return -5;
		if(0 > H5Pclose(plist_id)) return -6;
		if(0 > (transferPlist = H5Pcreate(H5P_DATASET_XFER))) return -7;
		if(0 > H5Pset_dxpl_mpio(transferPlist, (collectiveCalls) ? H5FD_MPIO_COLLECTIVE : H5FD_MPIO_INDEPENDENT)) return -8;
	}

	// Define the dataspace.
	floatCount = 1;
	for(i = 0; i < kDimensionCount; i++) {
		if(i) {
			dimensionMaxSizes[i] = kDimensionSizes[i];
			floatCount *= reducedDimensionSizes[i];
		} else {
			dimensionMaxSizes[i] = H5S_UNLIMITED;
		}
	}
	if(0 > (spaceId = H5Screate_simple(kDimensionCount, kDimensionSizes, dimensionMaxSizes))) return -9;
	if(0 > (datasetCreationPList = H5Pcreate (H5P_DATASET_CREATE))) return -10;
	if(0 > H5Pset_chunk(datasetCreationPList, kDimensionCount, kDimensionSizes)) return -11;

	//write the variable descriptions
	for(i = 0; i < kVariableCount; i++) {
		if(snprintf(name, sizeof(name), "variable %li", i+1) <= 0) return -12;
		if(0 > (setIds[i] = H5Dcreate(fileId, name, H5T_IEEE_F32BE, spaceId, H5P_DEFAULT, datasetCreationPList, H5P_DEFAULT))) return -13;
	}

	//fake some data
	float* data = (float*)malloc(sizeof(float)*floatCount);
	for(i = 0; i < floatCount; i++) data[i] = (float)i;

	// Setup some parameters.
	for(i = 0; i < kDimensionCount; i++) {
		startCoords[i] = 0;
		curDimensionSizes[i] = kDimensionSizes[i];
	}
	startCoords[1] = myStartCoord;
	curDimensionSizes[0] = 0;
	if(0 > (memspace = H5Screate_simple (kDimensionCount, reducedDimensionSizes, NULL))) return -14;

	//write it many times
	for(i = 0; i < kRecordCount; i++) {
		startCoords[0] = i;
		long curVar;
		curDimensionSizes[0]++;
		for(curVar = 0; curVar < kVariableCount; curVar++) {
			hid_t filespace;
			if(0 > H5Dextend(setIds[curVar], curDimensionSizes)) return -15;
			if(0 > (filespace = H5Dget_space(setIds[curVar]))) return -16;
			if(0 > H5Sselect_hyperslab(filespace, H5S_SELECT_SET, startCoords, NULL, reducedDimensionSizes, NULL)) return -17;
			if(0 > H5Dwrite(setIds[curVar], H5T_NATIVE_FLOAT, memspace, filespace, transferPlist, data)) return -18;
		}
	}
	
	// Cleanup.
	for(i = 0; i < kVariableCount; i++) if(0 > H5Dclose(setIds[i])) return -19;
	if(0 > H5Sclose(memspace)) return -20;
	if(0 > H5Sclose(spaceId)) return -21;
	if(0 > H5Pclose(transferPlist)) return -22;
	if(0 > H5Fclose(fileId)) return -23;

	// Now the interesting part: How long did we take?
	MPI_Barrier(MPI_COMM_WORLD);
	if(!myProcId) {
		double totalFloatCount = 1, elapsedTime, throughput;
		if(gettimeofday(&endTime, NULL)) return -19;
		for(i = 1; i < kDimensionCount; i++) totalFloatCount *= kDimensionSizes[i];
		elapsedTime = (double)(endTime.tv_sec - startTime.tv_sec) + (endTime.tv_usec - startTime.tv_usec)/(double)1000000;
		throughput = totalFloatCount*kRecordCount*kVariableCount*4/(1024*1024*elapsedTime);
		if(reducedOutput) {
			printf("%.3f\n", throughput);
		} else {
			printf("Elapsed time: %.6fs\n", elapsedTime);
			printf("Speed: %.3fMiByte/s\n", throughput);
		}
	}
	
	MPI_Finalize();
	
	return 0;
}
